/* prefs.c */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gconf/gconf-client.h>

#include "support.h"
#include "puzzle.h"
#include "interface.h"
#include "prefs.h"
#include "gfx.h"


#define PUZZLE_SIZE_MIN  3
#define PUZZLE_SIZE_MAX  5
#define DEFAULT_SIZE     4
#define DEFAULT_OVERLAY  OVERLAY_WHITE
#define DEFAULT_RANDOM   FALSE
#define DEFAULT_KEY_UP   GDK_Up
#define DEFAULT_KEY_DN   GDK_Down
#define DEFAULT_KEY_LT   GDK_Left
#define DEFAULT_KEY_RT   GDK_Right

extern Puzzle *puzzle;
extern GtkWidget *app;
extern gint image_px;

static GtkWidget *dialog_prefs = NULL;
static GtkWidget *button_move[4];
static GtkWidget *entry_move[4];
static GtkWidget *combo_size = NULL;
static GtkWidget *combo_image = NULL;

GConfClient *skoosh_gconf_client = NULL;
gchar       *fname_img_custom = NULL;
KeyID       key[4];
Overlay     overlay;
gint        img_id;



ImgFile img_file[] = {
	{
		NULL,
		N_("Custom"),
		0 /* (0 = allow any size) */
	},
	{
		PACKAGE_PIXMAP_DIR G_DIR_SEPARATOR_S "default.jpg",
		N_("Numbers"),
		4 /* (4 = restrict this theme to 4x4 tiles) */
	},
	{
		PACKAGE_PIXMAP_DIR G_DIR_SEPARATOR_S "griffs-summer-breeze.jpg",
		N_("Summer Breeze"),
		0
	}
	/* (add new images here) */
};
gint n_img_files = G_N_ELEMENTS(img_file);



static gint 
skoosh_gconf_get_int (gchar *key, gint default_int)
{
	GConfValue *value = NULL, *schema_value;
	gint retval = default_int;

	value = gconf_client_get (skoosh_gconf_client, key, NULL);
	if (value != NULL && value->type == GCONF_VALUE_INT) {
		retval = gconf_value_get_int (value);
		gconf_value_free (value);
	}
	else {
		schema_value = gconf_client_get_default_from_schema
		               (skoosh_gconf_client, key, NULL);
		if (schema_value != NULL) {
			retval = gconf_value_get_int (schema_value);
			gconf_value_free (schema_value);
		}
	}
	return retval;
}



static gboolean
skoosh_gconf_get_bool (gchar *key, gboolean default_bool)
{
	GConfValue *value = NULL, *schema_value;
	gboolean retval = default_bool;

	value = gconf_client_get (skoosh_gconf_client, key, NULL);
	if (value != NULL && value->type == GCONF_VALUE_BOOL) {
		retval = gconf_value_get_bool (value); 
		gconf_value_free (value);
	}
	else {
		schema_value = gconf_client_get_default_from_schema
		               (skoosh_gconf_client, key, NULL);
		if (schema_value != NULL) {
			retval = gconf_value_get_bool (schema_value);
			gconf_value_free (schema_value);
		}
	}
	return retval;
}



static gchar*
skoosh_gconf_get_string (gchar *key)
{
	GConfValue *value = NULL, *schema_value;
	gchar *retval = NULL;

	value = gconf_client_get (skoosh_gconf_client, key, NULL);
	if (value != NULL && value->type == GCONF_VALUE_STRING) {
		retval = g_strdup (gconf_value_get_string (value));
		gconf_value_free (value);
	}
	else {
		schema_value = gconf_client_get_default_from_schema
		               (skoosh_gconf_client, key, NULL);
		if (schema_value != NULL) {
			retval = g_strdup (gconf_value_get_string (schema_value));
			gconf_value_free (schema_value);
		}
	}
	return retval;
}



static void
prefs_load (void)
{
	gint size;

	fname_img_custom = skoosh_gconf_get_string ("/apps/skoosh/image");
	size = skoosh_gconf_get_int ("/apps/skoosh/size", DEFAULT_SIZE);
	overlay = skoosh_gconf_get_int ("/apps/skoosh/overlay", DEFAULT_OVERLAY);
	img_id = skoosh_gconf_get_int ("/apps/skoosh/img_id", IMG_ID_DEFAULT);
	puzzle_set_random (puzzle, skoosh_gconf_get_bool ("/apps/skoosh/random", DEFAULT_RANDOM));
	key[KEY_UP] = skoosh_gconf_get_int ("/apps/skoosh/keyup", DEFAULT_KEY_UP);
	key[KEY_DN] = skoosh_gconf_get_int ("/apps/skoosh/keydn", DEFAULT_KEY_DN);
	key[KEY_LT] = skoosh_gconf_get_int ("/apps/skoosh/keylt", DEFAULT_KEY_LT);
	key[KEY_RT] = skoosh_gconf_get_int ("/apps/skoosh/keyrt", DEFAULT_KEY_RT);

	/* sanity checks */
	if (size < PUZZLE_SIZE_MIN || size > PUZZLE_SIZE_MAX) {
		size = DEFAULT_SIZE;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/size", size, NULL);
	}
	if (overlay != OVERLAY_BLACK && overlay != OVERLAY_WHITE && overlay != OVERLAY_NONE) {
		overlay = DEFAULT_OVERLAY;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/overlay", overlay, NULL);
	}
	if ((img_id < 0 || img_id >= n_img_files) ||
	    (img_id == IMG_ID_CUSTOM && fname_img_custom == NULL)) {
		img_id = IMG_ID_DEFAULT;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/img_id", img_id, NULL);
	}
	if (fname_img_custom != NULL && strcmp (fname_img_custom, "none") == 0) {
		g_free (fname_img_custom);
		fname_img_custom = NULL;
	}

	puzzle_set_size (puzzle, size);
}


#if 0
static void
on_gconf_image_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_size_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_overlay_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_image_id_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_random_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_keyup_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_keydn_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_keylt_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
static void
on_gconf_keyrt_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
}
#endif


void
prefs_init (gint argc, gchar **argv)
{
	gconf_init (argc, argv, NULL);
	skoosh_gconf_client = gconf_client_get_default ();
	gconf_client_add_dir (skoosh_gconf_client, "/apps/skoosh",
	                      GCONF_CLIENT_PRELOAD_NONE, NULL);

	prefs_load ();

#if 0
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/image",
	                         on_gconf_image_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/size",
	                         on_gconf_size_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/overlay",
	                         on_gconf_overlay_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/img_id",
	                         on_gconf_image_id_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/random",
	                         on_gconf_random_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/keyup",
	                         on_gconf_keyup_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/keydn",
	                         on_gconf_keydn_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/keylt",
	                         on_gconf_keylt_changed, NULL, NULL, NULL);
	gconf_client_notify_add (skoosh_gconf_client, "/apps/skoosh/keyrt",
	                         on_gconf_keyrt_changed, NULL, NULL, NULL);
#endif
}



void
prefs_free (void)
{
	g_object_unref (G_OBJECT(skoosh_gconf_client));
	g_free (fname_img_custom);
}



static void
size_menu_set_sensitive (gboolean sensitive)
{
	if (combo_size == NULL) return;
	gtk_widget_set_sensitive (combo_size, sensitive);
}



void
prefs_set_image (gchar *fname)
{
	/* replace fname_img_custom and save to prefs */
	g_free (fname_img_custom);
	fname_img_custom = fname;
	gconf_client_set_string (skoosh_gconf_client, "/apps/skoosh/image",
	                         fname_img_custom, NULL);
}



void
prefs_set_image_id (gint id)
{
	/* replace img_id and update preferences */
	if (id != img_id) {
		img_id = id;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/img_id", img_id, NULL);
		if (img_file[img_id].size != 0) {
			/* this image requires a particular puzzle size */
			gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/size",
			                      img_file[img_id].size, NULL);
		}
	}

	/* update the prefs dialog */
	if (dialog_prefs != NULL) {
		gtk_combo_box_set_active (GTK_COMBO_BOX(combo_image), id);
		if (id == IMG_ID_CUSTOM) {
			/* FIXME: */
			/* gtk_widget_set_sensitive (menu_item_custom, TRUE); */
		}
		gtk_combo_box_set_active (GTK_COMBO_BOX(combo_size),
		                          puzzle_get_size (puzzle) - PUZZLE_SIZE_MIN);
		size_menu_set_sensitive (img_file[id].size == 0);
		gtk_combo_box_set_active (GTK_COMBO_BOX(combo_size),
		                          puzzle_get_size (puzzle) - PUZZLE_SIZE_MIN);
	}
}



static void
key_entries_set_sensitive (gboolean sensitive)
{
	gint i;

	for (i = 0; i < 4; i++) {
		gtk_widget_set_sensitive (entry_move[i], sensitive);
	}
}



static void
on_overlay_selected (GtkComboBox *combo, gpointer data)
{
	gint new_overlay = gtk_combo_box_get_active (combo);

	if (new_overlay != overlay) {
		overlay = new_overlay;
		gfx_create_tileset ();
		gfx_redraw_drawarea (TRUE);
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/overlay",
		                      overlay, NULL);
	}
}



static void
on_size_selected (GtkComboBox *combo, gpointer data)
{
	gint new_size;
   
	new_size = gtk_combo_box_get_active (combo) + PUZZLE_SIZE_MIN;
	if (new_size == puzzle_get_size (puzzle)) return;
	if (new_size < PUZZLE_SIZE_MIN || new_size > PUZZLE_SIZE_MAX) return;

	set_status (STATUS_CLEAR, NULL);

	puzzle_set_size (puzzle, new_size);
	gfx_create_tileset ();
	puzzle_display_tiles (puzzle);

	gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/size",
	                      puzzle_get_size (puzzle), NULL);
}



static void
on_image_selected (GtkComboBox *combo, gpointer data)
{
	gint id = gtk_combo_box_get_active (combo);

	if (id == IMG_ID_CUSTOM) {
		if (fname_img_custom != NULL) {
			if (!gfx_load_img_id (IMG_ID_CUSTOM)) {
				file_error_dialog (fname_img_custom, GTK_MESSAGE_WARNING);
				/* forget this image */
				/* FIXME: */
				/* gtk_widget_set_sensitive (menu_item_custom, FALSE); */
				prefs_set_image_id (img_id);
				gconf_client_unset (skoosh_gconf_client, "/apps/skoosh/image", NULL);
			}
		}
		else {
			prefs_set_image_id (IMG_ID_CUSTOM);
		}
	}
	else if (id != img_id) {
		if (!gfx_load_img_id (id)) {
			file_error_dialog (img_file[id].fname, GTK_MESSAGE_WARNING);
		}
	}
}



static void
on_button_browse_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *dialog;
	dialog = gtk_file_chooser_dialog_new (_("Open File"),
	             GTK_WINDOW(dialog_prefs),
	             GTK_FILE_CHOOSER_ACTION_OPEN,
	             GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
	             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
	gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);
	if (fname_img_custom != NULL) {
		/* use current file's location as starting point */
		gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(dialog),
		                               fname_img_custom);
	}
	if (gtk_dialog_run (GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)	{
		gchar *fname;
		fname = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
		if (fname != NULL && fname[0] != '\0') {
			if (gfx_load (fname, FALSE)) {
				prefs_set_image (g_strdup (fname));
				prefs_set_image_id (IMG_ID_CUSTOM);
			}
			else {
				file_error_dialog (fname, GTK_MESSAGE_WARNING);
				/* restore the original setting */
				prefs_set_image_id (img_id);
			}
		}
		g_free (fname);
	}
	gtk_widget_destroy (dialog);
}



#if PREFS_OVERKILL
static void
on_checkbutton_random_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	puzzle_set_random (puzzle, !GTK_TOGGLE_BUTTON(togglebutton)->active);
	gconf_client_set_bool (skoosh_gconf_client, "/apps/skoosh/random", IS_RANDOM(puzzle), NULL);
}
#endif



static void
on_button_move_clicked (GtkWidget *button, gpointer key_select)
{
	key_entries_set_sensitive (FALSE);
	gtk_widget_set_sensitive (entry_move[(gint)key_select], TRUE);
	gtk_widget_grab_focus (entry_move[(gint)key_select]);
}



static void
on_key_selected (GtkWidget *entry, GdkEventKey *data)
{
	gint key_select;

	gtk_entry_set_text (GTK_ENTRY(entry), gdk_keyval_name (data->keyval));

	if (entry == entry_move[KEY_UP]) {
		key_select = KEY_UP;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/keyup", data->keyval, NULL);
	}
	else if (entry == entry_move[KEY_DN]) {
		key_select = KEY_DN;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/keydn", data->keyval, NULL);
	}
	else if (entry == entry_move[KEY_LT]) {
		key_select = KEY_LT;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/keylt", data->keyval, NULL);
	}
	else {
		key_select = KEY_RT;
		gconf_client_set_int (skoosh_gconf_client, "/apps/skoosh/keyrt", data->keyval, NULL);
	}

	key[key_select] = data->keyval;
	gtk_widget_set_sensitive (entry, FALSE);
	gtk_widget_grab_focus (button_move[key_select]);
}



static const gchar*
theme_title (gint id)
{
	/* this lets gettext translate the theme menu */
	return img_file[id].title;
}



static void
on_prefs_response (GtkWidget *w, int response_id, gpointer data)
{
	if (response_id == GTK_RESPONSE_ACCEPT) {
		gtk_widget_hide (dialog_prefs);
	}
}



void
create_prefs_dialog (void)
{
	GtkWidget *notebook;
	GtkWidget *vbox_appearance;
	GtkWidget *table_appearance;
	GtkWidget *label_size;
	GtkWidget *label_decor;
	GtkWidget *label_image;
	GtkWidget *combo_decor;
	GtkWidget *button_browse;
#if PREFS_OVERKILL
	GtkWidget *checkbutton_random;
#endif
	GtkWidget *label_appearance;
	GtkWidget *table_keyboard;
	GtkWidget *label_keyboard;
	GtkTooltips *tooltips;
	gint i;


	if (dialog_prefs != NULL) {
		key_entries_set_sensitive (FALSE);
		gtk_window_present (GTK_WINDOW(dialog_prefs));
		return;
	}

	tooltips = gtk_tooltips_new ();

	dialog_prefs = gtk_dialog_new_with_buttons (_("Skoosh Preferences"),
	                   GTK_WINDOW (app),
	                   GTK_DIALOG_DESTROY_WITH_PARENT,
	                   GTK_STOCK_CLOSE, GTK_RESPONSE_ACCEPT,
	                   NULL);

	g_signal_connect (GTK_OBJECT(dialog_prefs), "destroy",
	                  GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog_prefs);

	g_signal_connect (dialog_prefs, "response",
	                  G_CALLBACK(on_prefs_response), &dialog_prefs);

	gtk_window_set_resizable (GTK_WINDOW(dialog_prefs), FALSE);

	notebook = gtk_notebook_new ();
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog_prefs)->vbox), notebook, TRUE, TRUE, 0);

	vbox_appearance = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (notebook), vbox_appearance);

	table_appearance = gtk_table_new (4, 2, FALSE);
	gtk_box_pack_start (GTK_BOX (vbox_appearance), table_appearance, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (table_appearance), 12);
	gtk_table_set_row_spacings (GTK_TABLE (table_appearance), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table_appearance), 12);

	label_size = gtk_label_new_with_mnemonic (_("_Puzzle Size:"));
	gtk_table_attach (GTK_TABLE (table_appearance), label_size, 0, 1, 0, 1,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_size), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_size), 0, 0.5);

	label_decor = gtk_label_new_with_mnemonic (_("_Tile Borders:"));
	gtk_table_attach (GTK_TABLE (table_appearance), label_decor, 0, 1, 1, 2,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_decor), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_decor), 0, 0.5);

	combo_decor = gtk_combo_box_new_text ();
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_decor), _("Black"));
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_decor), _("White"));
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_decor), _("None"));
	gtk_table_attach (GTK_TABLE (table_appearance), combo_decor, 1, 2, 1, 2,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (GTK_EXPAND), 0, 0);

	combo_size = gtk_combo_box_new_text ();
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_size), _("Three by Three"));
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_size), _("Four by Four"));
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo_size), _("Five by Five"));
	gtk_table_attach (GTK_TABLE (table_appearance), combo_size, 1, 2, 0, 1,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (GTK_EXPAND), 0, 0);

	combo_image = gtk_combo_box_new_text ();
	for (i = 0; i < n_img_files; i++) {
		gtk_combo_box_append_text (GTK_COMBO_BOX (combo_image), _(theme_title (i)));
	}
	/* FIXME: */
	/* gtk_widget_set_sensitive (, fname_img_custom != NULL); */
	gtk_table_attach (GTK_TABLE (table_appearance), combo_image, 1, 2, 2, 3,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	label_image = gtk_label_new_with_mnemonic (_("_Image:"));
	gtk_table_attach (GTK_TABLE (table_appearance), label_image, 0, 1, 2, 3,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_image), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_image), 0, 0.5);

	button_browse = gtk_button_new_with_mnemonic (_("_Browse..."));
	gtk_table_attach (GTK_TABLE (table_appearance), button_browse, 1, 2, 3, 4,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, button_browse,
	_("Alternatively, images can be dragged and dropped into the main window"), NULL);

#if PREFS_OVERKILL
	checkbutton_random = gtk_check_button_new_with_mnemonic (_("_Missing tile is always bottom-right"));
	gtk_box_pack_start (GTK_BOX (vbox_appearance), checkbutton_random, FALSE, FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (checkbutton_random), 12);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(checkbutton_random), !IS_RANDOM(puzzle));
#endif

	label_appearance = gtk_label_new_with_mnemonic (_("_Appearance"));
	gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 0), label_appearance);
	gtk_label_set_justify (GTK_LABEL (label_appearance), GTK_JUSTIFY_LEFT);

	gtk_label_set_mnemonic_widget (GTK_LABEL(label_size), combo_size);
	gtk_label_set_mnemonic_widget (GTK_LABEL(label_decor), combo_decor);
	gtk_label_set_mnemonic_widget (GTK_LABEL(label_image), combo_image);

	table_keyboard = gtk_table_new (4, 2, FALSE);
	gtk_container_add (GTK_CONTAINER (notebook), table_keyboard);
	gtk_container_set_border_width (GTK_CONTAINER (table_keyboard), 12);
	gtk_table_set_row_spacings (GTK_TABLE (table_keyboard), 0);
	gtk_table_set_col_spacings (GTK_TABLE (table_keyboard), 12);

	entry_move[KEY_UP] = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table_keyboard), entry_move[KEY_UP], 1, 2, 0, 1,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	entry_move[KEY_DN] = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table_keyboard), entry_move[KEY_DN], 1, 2, 1, 2,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	entry_move[KEY_LT] = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table_keyboard), entry_move[KEY_LT], 1, 2, 2, 3,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	entry_move[KEY_RT] = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table_keyboard), entry_move[KEY_RT], 1, 2, 3, 4,
	                  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	button_move[KEY_UP] = gtk_button_new_with_mnemonic (_("Move _Up"));
	gtk_table_attach (GTK_TABLE (table_keyboard), button_move[KEY_UP], 0, 1, 0, 1,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	button_move[KEY_DN] = gtk_button_new_with_mnemonic (_("Move _Down"));
	gtk_table_attach (GTK_TABLE (table_keyboard), button_move[KEY_DN], 0, 1, 1, 2,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	button_move[KEY_LT] = gtk_button_new_with_mnemonic (_("Move _Left"));
	gtk_table_attach (GTK_TABLE (table_keyboard), button_move[KEY_LT], 0, 1, 2, 3,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	button_move[KEY_RT] = gtk_button_new_with_mnemonic (_("Move _Right"));
	gtk_table_attach (GTK_TABLE (table_keyboard), button_move[KEY_RT], 0, 1, 3, 4,
	                  (GtkAttachOptions) (GTK_FILL),
	                  (GtkAttachOptions) (0), 0, 0);

	label_keyboard = gtk_label_new_with_mnemonic (_("_Keyboard"));
	gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1), label_keyboard);
	gtk_label_set_justify (GTK_LABEL (label_keyboard), GTK_JUSTIFY_LEFT);

	g_signal_connect (GTK_OBJECT (button_browse), "clicked",
	                  GTK_SIGNAL_FUNC (on_button_browse_clicked), NULL);

#if PREFS_OVERKILL
	g_signal_connect (GTK_OBJECT (checkbutton_random), "toggled",
	                  GTK_SIGNAL_FUNC (on_checkbutton_random_toggled), NULL);
#endif

	gtk_combo_box_set_active (GTK_COMBO_BOX(combo_decor), overlay);
	gtk_combo_box_set_active (GTK_COMBO_BOX(combo_size), puzzle_get_size (puzzle) - PUZZLE_SIZE_MIN);
	gtk_combo_box_set_active (GTK_COMBO_BOX(combo_image), img_id);
	size_menu_set_sensitive (img_file[img_id].size == 0);

	g_signal_connect (G_OBJECT (combo_decor), "changed",
	                  G_CALLBACK (on_overlay_selected), NULL);
	g_signal_connect (G_OBJECT (combo_size), "changed",
	                  G_CALLBACK (on_size_selected), NULL);
	g_signal_connect (G_OBJECT (combo_image), "changed",
	                  G_CALLBACK (on_image_selected), NULL);

	key_entries_set_sensitive (FALSE);
	for (i = 0; i < 4; i++) {
		gtk_entry_set_text (GTK_ENTRY(entry_move[i]), gdk_keyval_name (key[i]));
		gtk_editable_set_editable (GTK_EDITABLE(entry_move[i]), FALSE);
		g_signal_connect (GTK_OBJECT(button_move[i]), "clicked",
		                  GTK_SIGNAL_FUNC(on_button_move_clicked), (gpointer)i);
		g_signal_connect (GTK_OBJECT(entry_move[i]), "key_press_event",
		                  GTK_SIGNAL_FUNC(on_key_selected), NULL);
	}

	gtk_widget_show_all (dialog_prefs);
}

